#!/bin/bash
########################################
#
#      This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 3 of the License, or
#      (at your option) any later version.
#      
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################
#
# Wizard to get options for gsecraif and invoke gsecraif to split
# an original file in to components
#
# $Date$
# $Revision$
#
########################################
file2split=""
targetdir=`pwd`
numoffiles=3
useprefix=y
prefix="split"
defaultprefix=y
rotate=3
debuglevel=0
hash=n
compfilename=()
argline=""	# arguments to pass to gsecriaf

function getfilename {
    local isvalidfile=0
    while [ $isvalidfile = 0 ]; do
        echo "Enter the full path and file name of the file to split:"
        read file2split
        if [ -f $file2split ]; then
            isvalidfile=1
        else
            isvalidfile=0
            echo "${filetosplit} can not be found or is not a file."
        fi
    done
}

function gettargetdir {
    local isvaliddir=0
    while [ $isvaliddir = 0 ]; do
        echo "Enter the full path and directory that will contain the components:"
        read targetdir
        if [ -d $targetdir ]; then
            isvaliddir=1
        else
            isvaliddir=0
            echo "${targetdir} can not be found or is not a directory."
        fi
        if [ -w $targetdir ]; then
            isvaliddir=1
        else
            isvaliddir=0
            echo "${targetdir} does not have write permission."
        fi
    done
}

function getnumfiles {
    local validnum=0
    local tempnum
    while [ $validnum = 0 ]; do
        echo "Enter the number of component files (3-255) [${numoffiles}]"
        read tempnum
        if [ -z $tempnum ]; then
            tempnum=$numoffiles
        fi
        if [ $(($tempnum)) -gt 2 ] && [ $(($tempnum)) -lt 256 ]; then
            validnum=1
            numoffiles=$tempnum
        fi
    done
}

function prefixflag {
    echo "Is a prefix to be used for the component file names? [y]"
    read useprefix
    if [ -z $useprefix ]; then
        useprefix="y"
    fi
    if [ $useprefix = y ] || [ $useprefix = Y ]; then
        useprefix="y"
    else
        useprefix="n"
    fi
}

function setprefix {
    echo "Enter required prefix (or just Enter to use default) []"
    read prefix
    if [ -z $prefix ]; then
        defaultprefix="y"
        prefix="split"
    else 
        defaultprefix="n"
    fi
}

function getfilenames {
    local fileid=0
    while [ $fileid -lt $numoffiles ]; do
        echo "Enter component file name ${fileid}"
        read compfilename[${fileid}]
        let fileid=${fileid}+1
    done
}

function hashflag {
    echo "Print hash symbols as data is processed? [n]"
    read hash
    if [ -z $hash ]; then
        hash="n"
    fi
    if [ $hash = y ] || [ $hash = Y ]; then
        hash="y"
    else
        hash="n"
    fi
}
 
function getrotate {
    local validnum=0
    local tempnum
    while [ $validnum = 0 ]; do
        echo "Enter a value for the number of bits to rotate (0-7) [${rotate}]"
        read tempnum
        if [ -z $tempnum ]; then
            tempnum=$rotate
        fi
        if [ $(($temnum)) -gt -1 ] && [ $(($tempnum)) -lt 9 ]; then
            validnum=1
            rotate=$tempnum
        fi
    done
} 

function getdebuglevel {
    local validnum=0
    local tempnum
    while [ $validnum = 0 ]; do
        echo "Enter debug level (0-3) [${debuglevel}]"
        read tempnum
        if [ -z $tempnum ]; then
            tempnum=$debuglevel
        fi
        if [ $(($temnum)) -gt -1 ] && [ $(($tempnum)) -lt 4 ]; then
            validnum=1
            debuglevel=$tempnum
        fi
    done
}

# Main section:

echo "Script to split file with gsecraif."
getfilename
getnumfiles
prefixflag
if [ "$useprefix" = "y" ]; then
    setprefix
    gettargetdir
else
    getfilenames
fi
getrotate
hashflag
getdebuglevel
# build the argument line:

if [ "$useprefix" = "y" ]; then
    argline="${argline} -n ${numoffiles} -p ${prefix}"
fi
if [ $debuglevel -gt 0 ]; then
    argline="${argline} -d ${debuglevel}"
fi
argline="${argline} -r ${rotate}"
if [ $hash = y ]; then
    argline="${argline} -#"
fi
if [ "$useprefix" != "y" ]; then
    fileid=0
    while [ $fileid -lt $numoffiles ]; do
        echo "${compfilename[${fileid}]}"
        argline="${argline} ${compfilename[${fileid}]}"
        echo "${argline}"
        let fileid=${fileid}+1
    done
fi

echo
echo "Name of file to split: ${file2split}"
echo "Number of files: ${numoffiles}"
echo "Use prefix: ${useprefix}"
if [ "$useprefix" = "y" ]; then
    echo "Prefix: ${prefix}"
    echo "Directory to contain components: ${targetdir}"
fi
echo "Rotate: ${rotate}"
echo "Debug level: ${debuglevel}"
echo "Show hashes: ${hash}"
echo "Argument line: ${argline}"
echo
echo "Split ${file2split} [y]"
read proceed
if [ -z $proceed ]; then
    proceed="y"
fi
if [ $proceed = y ] || [ $proceed = Y ]; then
    cd ${targetdir}
    cat ${file2split} | gsecraif ${argline}
fi
exit 0
